@@ -1,27 +1,13 @@
-/*=============================================================================
-
-    This file is part of ARB.
-
-    ARB is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    ARB is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with ARB; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
-
-=============================================================================*/
-/******************************************************************************
-
+/*
     Copyright (C) 2012 Fredrik Johansson
 
-******************************************************************************/
+    This file is part of Arb.
+
+    Arb is free software: you can redistribute it and/or modify it under
+    the terms of the GNU Lesser General Public License (LGPL) as published
+    by the Free Software Foundation; either version 2.1 of the License, or
+    (at your option) any later version.  See <http://www.gnu.org/licenses/>.
+*/
 
 #ifndef ACB_MAT_H
 #define ACB_MAT_H
@@ -32,10 +18,11 @@
 #define ACB_MAT_INLINE static __inline__
 #endif
 
-#include "arb.h"
-#include "acb.h"
+#include <stdio.h>
 #include "flint/fmpz_mat.h"
 #include "flint/fmpq_mat.h"
+#include "arb.h"
+#include "acb.h"
 #include "arb_mat.h"
 #include "acb_poly.h"
 
@@ -79,6 +66,16 @@ acb_mat_swap(acb_mat_t mat1, acb_mat_t mat2)
     *mat2 = t;
 }
 
+/* Window matrices */
+
+void acb_mat_window_init(acb_mat_t window, const acb_mat_t mat, slong r1, slong c1, slong r2, slong c2);
+
+ARB_MAT_INLINE void
+acb_mat_window_clear(acb_mat_t window)
+{
+    flint_free(window->rows);
+}
+
 /* Conversions */
 
 void acb_mat_set(acb_mat_t dest, const acb_mat_t src);
@@ -99,7 +96,13 @@ void acb_mat_randtest(acb_mat_t mat, flint_rand_t state, slong prec, slong mag_b
 
 /* I/O */
 
-void acb_mat_printd(const acb_mat_t mat, slong digits);
+void acb_mat_fprintd(FILE * file, const acb_mat_t mat, slong digits);
+
+ACB_MAT_INLINE void
+acb_mat_printd(const acb_mat_t mat, slong digits)
+{
+    acb_mat_fprintd(stdout, mat, digits);
+}
 
 /* Comparisons */
 
@@ -119,18 +122,74 @@ int acb_mat_contains_fmpz_mat(const acb_mat_t mat1, const fmpz_mat_t mat2);
 
 int acb_mat_is_real(const acb_mat_t mat);
 
+ACB_MAT_INLINE int
+acb_mat_is_empty(const acb_mat_t mat)
+{
+    return (mat->r == 0) || (mat->c == 0);
+}
+
+ACB_MAT_INLINE int
+acb_mat_is_square(const acb_mat_t mat)
+{
+    return (mat->r == mat->c);
+}
+
+/* Radius and interval operations */
+
+ACB_MAT_INLINE void
+acb_mat_get_mid(acb_mat_t B, const acb_mat_t A)
+{
+    slong i, j;
+
+    for (i = 0; i < acb_mat_nrows(A); i++)
+    {
+        for (j = 0; j < acb_mat_ncols(A); j++)
+        {
+            arb_get_mid_arb(acb_realref(acb_mat_entry(B, i, j)), acb_realref(acb_mat_entry(A, i, j)));
+            arb_get_mid_arb(acb_imagref(acb_mat_entry(B, i, j)), acb_imagref(acb_mat_entry(A, i, j)));
+        }
+    }
+}
+
+ACB_MAT_INLINE void
+acb_mat_add_error_mag(acb_mat_t mat, const mag_t err)
+{
+    slong i, j;
+
+    for (i = 0; i < acb_mat_nrows(mat); i++)
+        for (j = 0; j < acb_mat_ncols(mat); j++)
+            acb_add_error_mag(acb_mat_entry(mat, i, j), err);
+}
+
 /* Special matrices */
 
 void acb_mat_zero(acb_mat_t mat);
 
 void acb_mat_one(acb_mat_t mat);
 
+void acb_mat_ones(acb_mat_t mat);
+
+void acb_mat_dft(acb_mat_t res, int kind, slong prec);
+
 void acb_mat_transpose(acb_mat_t mat1, const acb_mat_t mat2);
 
+void acb_mat_conjugate(acb_mat_t mat1, const acb_mat_t mat2);
+
+ACB_MAT_INLINE void
+acb_mat_conjugate_transpose(acb_mat_t mat1, const acb_mat_t mat2)
+{
+    acb_mat_transpose(mat1, mat2);
+    acb_mat_conjugate(mat1, mat1);
+}
+
 /* Norms */
 
 void acb_mat_bound_inf_norm(mag_t b, const acb_mat_t A);
 
+void acb_mat_frobenius_norm(arb_t res, const acb_mat_t A, slong prec);
+
+void acb_mat_bound_frobenius_norm(mag_t b, const acb_mat_t A);
+
 /* Arithmetic */
 
 void acb_mat_neg(acb_mat_t dest, const acb_mat_t src);
@@ -139,8 +198,14 @@ void acb_mat_add(acb_mat_t res, const acb_mat_t mat1, const acb_mat_t mat2, slon
 
 void acb_mat_sub(acb_mat_t res, const acb_mat_t mat1, const acb_mat_t mat2, slong prec);
 
+void acb_mat_mul_classical(acb_mat_t res, const acb_mat_t mat1, const acb_mat_t mat2, slong prec);
+void acb_mat_mul_threaded(acb_mat_t res, const acb_mat_t mat1, const acb_mat_t mat2, slong prec);
+void acb_mat_mul_reorder(acb_mat_t res, const acb_mat_t mat1, const acb_mat_t mat2, slong prec);
 void acb_mat_mul(acb_mat_t res, const acb_mat_t mat1, const acb_mat_t mat2, slong prec);
 
+void acb_mat_mul_entrywise(acb_mat_t res, const acb_mat_t mat1, const acb_mat_t mat2, slong prec);
+
+void acb_mat_sqr_classical(acb_mat_t res, const acb_mat_t mat, slong prec);
 void acb_mat_sqr(acb_mat_t res, const acb_mat_t mat, slong prec);
 
 void acb_mat_pow_ui(acb_mat_t B, const acb_mat_t A, ulong exp, slong prec);
@@ -303,19 +368,44 @@ acb_mat_swap_rows(acb_mat_t mat, slong * perm, slong r, slong s)
 slong acb_mat_find_pivot_partial(const acb_mat_t mat,
                                     slong start_row, slong end_row, slong c);
 
+void acb_mat_solve_tril_classical(acb_mat_t X, const acb_mat_t L, const acb_mat_t B, int unit, slong prec);
+void acb_mat_solve_tril_recursive(acb_mat_t X, const acb_mat_t L, const acb_mat_t B, int unit, slong prec);
+void acb_mat_solve_tril(acb_mat_t X, const acb_mat_t L, const acb_mat_t B, int unit, slong prec);
+
+void acb_mat_solve_triu_classical(acb_mat_t X, const acb_mat_t U, const acb_mat_t B, int unit, slong prec);
+void acb_mat_solve_triu_recursive(acb_mat_t X, const acb_mat_t U, const acb_mat_t B, int unit, slong prec);
+void acb_mat_solve_triu(acb_mat_t X, const acb_mat_t U, const acb_mat_t B, int unit, slong prec);
+
+int acb_mat_lu_classical(slong * P, acb_mat_t LU, const acb_mat_t A, slong prec);
+int acb_mat_lu_recursive(slong * P, acb_mat_t LU, const acb_mat_t A, slong prec);
 int acb_mat_lu(slong * P, acb_mat_t LU, const acb_mat_t A, slong prec);
 
 void acb_mat_solve_lu_precomp(acb_mat_t X, const slong * perm,
     const acb_mat_t A, const acb_mat_t B, slong prec);
 
+int acb_mat_solve_lu(acb_mat_t X, const acb_mat_t A, const acb_mat_t B, slong prec);
+
 int acb_mat_solve(acb_mat_t X, const acb_mat_t A, const acb_mat_t B, slong prec);
 
+int acb_mat_solve_precond(acb_mat_t X, const acb_mat_t A, const acb_mat_t B, slong prec);
+
+void acb_mat_approx_mul(acb_mat_t C, const acb_mat_t A, const acb_mat_t B, slong prec);
+void acb_mat_approx_solve_triu(acb_mat_t X, const acb_mat_t U, const acb_mat_t B, int unit, slong prec);
+void acb_mat_approx_solve_tril(acb_mat_t X, const acb_mat_t L, const acb_mat_t B, int unit, slong prec);
+int acb_mat_approx_lu(slong * P, acb_mat_t LU, const acb_mat_t A, slong prec);
+void acb_mat_approx_solve_lu_precomp(acb_mat_t X, const slong * perm, const acb_mat_t A, const acb_mat_t B, slong prec);
+int acb_mat_approx_solve(acb_mat_t X, const acb_mat_t A, const acb_mat_t B, slong prec);
+
 int acb_mat_inv(acb_mat_t X, const acb_mat_t A, slong prec);
 
+void acb_mat_det_lu(acb_t det, const acb_mat_t A, slong prec);
+void acb_mat_det_precond(acb_t det, const acb_mat_t A, slong prec);
 void acb_mat_det(acb_t det, const acb_mat_t A, slong prec);
 
 /* Special functions */
 
+void acb_mat_exp_taylor_sum(acb_mat_t S, const acb_mat_t A, slong N, slong prec);
+
 void acb_mat_exp(acb_mat_t B, const acb_mat_t A, slong prec);
 
 void _acb_mat_charpoly(acb_ptr cp, const acb_mat_t mat, slong prec);
@@ -324,6 +414,12 @@ void acb_mat_charpoly(acb_poly_t cp, const acb_mat_t mat, slong prec);
 
 void acb_mat_trace(acb_t trace, const acb_mat_t mat, slong prec);
 
+ACB_MAT_INLINE slong
+acb_mat_allocated_bytes(const acb_mat_t x)
+{
+    return _acb_vec_allocated_bytes(x->entries, x->r * x->c) + x->r * sizeof(acb_ptr);
+}
+
 #ifdef __cplusplus
 }
 #endif
